#include "Graphics.h"

#include <assert.h>
#include <mmsystem.h>
#include <string>

// A structure for our custom vertex
struct TCubeVertex
{
	D3DXVECTOR3 position; // The 3D position for the vertex
	D3DXVECTOR3 normal;   // The surface normal for the vertex
};

// Our custom FVF, which describes our custom vertex structure above
#define D3DFVF_CUBEVERTEX (D3DFVF_XYZ|D3DFVF_NORMAL)

// Constructor
CGraphics::CGraphics(void) : m_d3dObject(0),m_d3dDevice(0),m_vertexBuffer(0),
	m_indexBuffer(0)
{
}

// Destructor
CGraphics::~CGraphics(void)
{
	// Once we have finished with the Direct3D objects we must release them
	if (m_vertexBuffer)
		m_vertexBuffer->Release();

	if (m_indexBuffer)
		m_indexBuffer->Release();

	if (m_d3dDevice)
		m_d3dDevice->Release();

	if (m_d3dObject)
		m_d3dObject->Release();
}

/*
It is important, to aid debugging, to check the return code from DirectX calls
DirectX also provides a means of getting the error code in terms of an error string and
description. To use this we need to include dxerr.h and link with dxerr.lib. I find it
convenient to wrap this in this helper function.
Note: the function returns false if there was an error
*/
bool CGraphics::CheckHR(HRESULT hr)
{
	if (SUCCEEDED(hr))
		return true;

	// An error occured so find out what it was and output it to the output pane in Vis
	OutputDebugString("\nDirectX Error Reported:\n");
	OutputDebugString(DXGetErrorString(hr));
	OutputDebugString(" : ");
	OutputDebugString(DXGetErrorDescription(hr));
	OutputDebugString("\n\n");

	return false;
}

/* 
Initialise our graphics
If anything goes wrong return false
*/
bool CGraphics::Initialise(HWND hWnd)
{
	if (!InitialiseDirect3D(hWnd))
		return false;

	if (!SetupLighting())
		return false;

	if (!SetupMatrices())
		return false;

	if (!CreateCube())
		return false;

	return true;
}

/* 
Initialise Direct3D by creating the Direct3D object which we then use to create a Direct3D device
that represents the video card. 
*/
bool CGraphics::InitialiseDirect3D(HWND hWnd)
{
	// Create the main D3D object. It is only used to get a device object.
	m_d3dObject=Direct3DCreate9(D3D_SDK_VERSION);
	if (!m_d3dObject)
		return false;

	/* When creating our device we can specify a number of parameters describing our required device
	Since this demo runs in a window we are limited to the display depth and size etc. set by the user
	This means we only fill in a few parameters (we want a z buffer, run in a window). If you were
	wishing to run full screen you would need to fill in many more. See the site for details.*/
	D3DPRESENT_PARAMETERS presParams;
	ZeroMemory(&presParams,sizeof(presParams));
	presParams.Windowed=TRUE;
	presParams.SwapEffect=D3DSWAPEFFECT_DISCARD;
	presParams.BackBufferFormat=D3DFMT_UNKNOWN;
	presParams.EnableAutoDepthStencil = TRUE;
	presParams.AutoDepthStencilFormat = D3DFMT_D16;

	/*
	Once we have filled in our structure we can ask Direct3D to create us a device matching our requirements
	This call also takes some further parameters including the type of vertex processing support we need
	Since some computers do not support hardware vertex processing in a window I back off if hardware vertex
	processing is not available and instead try mixed vertex processing. If that fails I try software vertex
	processing which is obviously slower but supported by more machines. If all these fail we cannot create a
	device capable of running the demo.
	*/
	HRESULT hr=m_d3dObject->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd,D3DCREATE_HARDWARE_VERTEXPROCESSING,
		&presParams, &m_d3dDevice);
	if (FAILED(hr))
	{
		hr=m_d3dObject->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd,D3DCREATE_MIXED_VERTEXPROCESSING,
			&presParams, &m_d3dDevice);
		if (FAILED(hr))
		{
			hr=m_d3dObject->CreateDevice(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,hWnd,D3DCREATE_SOFTWARE_VERTEXPROCESSING,
				&presParams, &m_d3dDevice);
			if (FAILED(hr))
				return false;
		}
	}

	// As well as specifying that we want a z buffer (above) we also need to turn it on
	m_d3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

	return true;
}

/*
Since our cube is defined as having a normal (see FVF definition at the top of this file) it will be automatically
rendered using lighting. Therefore we need to turn some lights on before we can see it! It is possible to render
3D geometry without lighting but to do that we would have had to have included a diffuse colour component in our
FVF. Please see the site notes for details of the different types of lighting and how lighting and materials
work together
*/
bool CGraphics::SetupLighting()
{
	/* Set up a material. The material here just has the diffuse and ambient
	colors set to yellow. This means it will reflect just the red and green
	components of the directional and ambient lighting respectively. */
	ZeroMemory( &m_material, sizeof(D3DMATERIAL9) );
	m_material.Diffuse.r = m_material.Ambient.r = 1.0f;
	m_material.Diffuse.g = m_material.Ambient.g = 1.0f;
	m_material.Diffuse.b = m_material.Ambient.b = 0.0f;
	m_material.Diffuse.a = m_material.Ambient.a = 1.0f;    

	/* Set up a white, directional light. For details please see the notes on the site
	Note that many lights may be active at a time (normally limited by th graphic card to 8)*/
	D3DXVECTOR3 vecDir;
	D3DLIGHT9 light;
	ZeroMemory( &light, sizeof(D3DLIGHT9) );
	light.Type       = D3DLIGHT_DIRECTIONAL;
	light.Diffuse.r  = 1.0f;
	light.Diffuse.g  = 1.0f;
	light.Diffuse.b  = 1.0f;

	vecDir = D3DXVECTOR3(0.0f,0.0f,1.0f);    
	D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );

	light.Range = 1000.0f;
	if (!CheckHR(m_d3dDevice->SetLight( 0, &light )))
		return false;

	if (!CheckHR(m_d3dDevice->LightEnable( 0, TRUE )))
		return false;

	// Also turn on some ambient light.
	m_d3dDevice->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_XRGB(20,20,20));

	// Turn on the lighting generally
	m_d3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );

	return true;
}

/*
To render anything in Direct3D we need to set the three matrices
World Matrix - used to transform our geometry from object space into world space
i.e. position our geometry in the game world
View Matrix - used to transform the geometry form world space into view space
i.e. position the viewer (camera) in the world
Projection Matrix - transforms from view space into screen space
i.e. applies perspective projection to create the final 2D view
*/
bool CGraphics::SetupMatrices()
{
	/* For the projection matrix, we set up a perspective transform (which
	transforms geometry from 3D view space to 2D viewport space, with
	a perspective divide making objects smaller in the distance). To build
	a perpsective transform, we need the field of view (1/4 pi is common),
	the aspect ratio, and the near and far clipping planes (which define at
	what distances geometry should no longer be rendered).
	Note: we normally only set this matrix once */
	float aspectRation = 800.0f / 600.0f;

	D3DXMATRIX matProj;
	D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, aspectRation, 1.0f, 100.0f );

	m_d3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

	// Now for our view matrix. A view matrix can be defined given an eye point,
	// a point to lookat, and a direction for which way is up. Here, we set the
	// eye 10 units back along the z-axis, look at the origin, 
	// and define "up" to be in the y-direction.
	// We normally set this matrix whenever the camera changes position or orientation
	D3DXVECTOR3 vEyePt( 0.0f, 0.0f,-10.0f );
	D3DXVECTOR3 vLookatPt( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3 vUpVec( 0.0f, 1.0f, 0.0f );
	D3DXMATRIX matView;
	D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );

	m_d3dDevice->SetTransform( D3DTS_VIEW, &matView );

	// The world matrix is set just before rendering our cube
	return true;
}

/*
Create our cube
A cube has 6 sides which we can make up of two triangles per side, meaning we have 12 triangles
We cannot share vertices when drawing our cube as we need to have different normals 
for each vertex depending on which side the vertex is on, therefore we will need to have 4 
vertices per side which means 6*4 = 24 vertices in all
*/
bool CGraphics::CreateCube()
{
	/* We create a vertex buffer to hold our vertices via a function call of the device
	We need to specify how big the buffer should be in bytes, the FVF used and some flags
	If all goes well the call will create a vertex buffer object and point our pointer at it */
	HRESULT hr=m_d3dDevice->CreateVertexBuffer( 4*6*sizeof(TCubeVertex),D3DUSAGE_WRITEONLY,
		D3DFVF_CUBEVERTEX,D3DPOOL_DEFAULT, &m_vertexBuffer, NULL );

	if (!CheckHR(hr))
		return false;

	/* Fill the vertex buffer. 
	I find it best to first define the corners of the cube in two temporary arrays
	called front and back. I then use these when assigning positions to the vertex
	It can be quite complex to imagine the shape so drawing it can be useful and in
	fact I found for this taks that using an old CD box and labelling the corners
	was very useful! The main thing to remember is that when looking at a face the 
	vertices for each triangle must be defined in clockwise order*/

	// Note: the centre is 0,0,0 and the extents from -1,-1,-1 to +1,+1,+1

	// Front
	D3DXVECTOR3 front[4];	
	front[0]=D3DXVECTOR3(-1.0f,-1.0f,-1.0f);
	front[1]=D3DXVECTOR3(-1.0f,1.0f,-1.0f);
	front[2]=D3DXVECTOR3(1.0f,1.0f,-1.0f);
	front[3]=D3DXVECTOR3(1.0f,-1.0f,-1.0f);

	// Back
	D3DXVECTOR3 back[4];	
	back[0]=D3DXVECTOR3(1.0f,-1.0f,1.0f);
	back[1]=D3DXVECTOR3(1.0f,1.0f,1.0f);
	back[2]=D3DXVECTOR3(-1.0f,1.0f,1.0f);
	back[3]=D3DXVECTOR3(-1.0f,-1.0f,1.0f);

	/*In order to write to our vertex buffer we need to first lock it and get access to the memory	
	This is needed because the data could very well be held on the graphics card itself (for speed)
	and so to access it we need to lock. Note: try to avoid doing this as little as possible during 
	the render loop as there is a performance hit*/
	TCubeVertex* pVertices;
	hr=m_vertexBuffer->Lock( 0, 0, (void**)&pVertices, 0 );
	if (!CheckHR(hr))
		return false;

	// Now define our cube faces

	// Front facing, note: normals face away from face
	pVertices[0].position=front[0];
	pVertices[0].normal=D3DXVECTOR3(0.0f,0.0f,-1.0f);
	pVertices[1].position=front[1];
	pVertices[1].normal=D3DXVECTOR3(0.0f,0.0f,-1.0f);
	pVertices[2].position=front[2];
	pVertices[2].normal=D3DXVECTOR3(0.0f,0.0f,-1.0f);
	pVertices[3].position=front[3];
	pVertices[3].normal=D3DXVECTOR3(0.0f,0.0f,-1.0f);

	// Left face, remember to define in clockwise fashion (if looking at face)
	pVertices[4].position=back[3];
	pVertices[4].normal=D3DXVECTOR3(-1.0f,0.0f,0.0f);
	pVertices[5].position=back[2];
	pVertices[5].normal=D3DXVECTOR3(-1.0f,0.0f,0.0f);
	pVertices[6].position=front[1];
	pVertices[6].normal=D3DXVECTOR3(-1.0f,0.0f,0.0f);
	pVertices[7].position=front[0];
	pVertices[7].normal=D3DXVECTOR3(-1.0f,0.0f,0.0f);

	// Back face
	pVertices[8].position=back[0];
	pVertices[8].normal=D3DXVECTOR3(0.0f,0.0f,1.0f);
	pVertices[9].position=back[1];
	pVertices[9].normal=D3DXVECTOR3(0.0f,0.0f,1.0f);
	pVertices[10].position=back[2];
	pVertices[10].normal=D3DXVECTOR3(0.0f,0.0f,1.0f);
	pVertices[11].position=back[3];
	pVertices[11].normal=D3DXVECTOR3(0.0f,0.0f,1.0f);

	// Right face
	pVertices[12].position=front[3];
	pVertices[12].normal=D3DXVECTOR3(1.0f,0.0f,0.0f);
	pVertices[13].position=front[2];
	pVertices[13].normal=D3DXVECTOR3(1.0f,0.0f,0.0f);
	pVertices[14].position=back[1];
	pVertices[14].normal=D3DXVECTOR3(1.0f,0.0f,0.0f);
	pVertices[15].position=back[0];
	pVertices[15].normal=D3DXVECTOR3(1.0f,0.0f,0.0f);

	// Top face
	pVertices[16].position=front[1];
	pVertices[16].normal=D3DXVECTOR3(0.0f,1.0f,0.0f);
	pVertices[17].position=back[2];
	pVertices[17].normal=D3DXVECTOR3(0.0f,1.0f,0.0f);
	pVertices[18].position=back[1];
	pVertices[18].normal=D3DXVECTOR3(0.0f,1.0f,0.0f);
	pVertices[19].position=front[2];
	pVertices[19].normal=D3DXVECTOR3(0.0f,1.0f,0.0f);

	// Bottom face
	pVertices[20].position=back[3];
	pVertices[20].normal=D3DXVECTOR3(0.0f,-1.0f,0.0f);
	pVertices[21].position=front[0];
	pVertices[21].normal=D3DXVECTOR3(0.0f,-1.0f,0.0f);
	pVertices[22].position=front[3];
	pVertices[22].normal=D3DXVECTOR3(0.0f,-1.0f,0.0f);
	pVertices[23].position=back[0];
	pVertices[23].normal=D3DXVECTOR3(0.0f,-1.0f,0.0f);

	// Once finished we MUST unlock the vertex buffer
	m_vertexBuffer->Unlock();

	/* We now have vertices for each 4 corners of each face
	Each face will have two triangles defined in a clockwise order
	We need an index buffer to specify which vertices make up which triangles
	We will need 3 entries per triangle, 2 triangles per face and we have 6 faces
	giving 36 indices all

	We create an index buffer to hold all the indices defining our triangles via a 
	function call of the device. We need to specify how big the buffer should be in bytes, 
	along with some flags (see site for details).
	If all goes well the call will create an index buffer object and point our pointer at it */	
	hr=m_d3dDevice->CreateIndexBuffer(36* sizeof(WORD),D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_MANAGED,&m_indexBuffer,NULL);
	if (!CheckHR(hr))
		return false;


	// Lock the index buffer to access the data
	WORD *indices=NULL;
	hr=m_indexBuffer->Lock( 0, 0, (void**)&indices, 0 );
	if (!CheckHR(hr))
		return false;

	// Fill in the indices

	// Front face
	WORD indexStart=0;
	int count=0;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;

	// Left face
	indexStart+=4;
	count+=6;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;

	// Back face
	indexStart+=4;
	count+=6;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;

	// Right face
	indexStart+=4;
	count+=6;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;

	// Top face
	indexStart+=4;
	count+=6;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;

	// Bottom face
	indexStart+=4;
	count+=6;
	indices[count+0]=indexStart+0;indices[count+1]=indexStart+1;indices[count+2]=indexStart+2;
	indices[count+3]=indexStart+0;indices[count+4]=indexStart+2;indices[count+5]=indexStart+3;
	count+=6;

	// Sanity check!
	assert(count==36);

	// and again remember to unlock the buffer
	m_indexBuffer->Unlock();

	return true;
}

/*
When rendering we always follow the same steps
1. Clear the screen (target) and z buffer
2. Call BeginScene to tell D3D we are about to draw
3. Render all our geometry
4. Call EndScene and Present to show the new scene
*/
bool CGraphics::Render()
{
	// 1. clear the back buffer to black (0,0,0) and the z buffer to 1
	HRESULT hr=m_d3dDevice->Clear(0,NULL,D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,D3DCOLOR_XRGB(0,0,0),1.0f,0);
	if (!CheckHR(hr))
		return false;

	// 2. begin the scene
	hr=m_d3dDevice->BeginScene();
	if (!CheckHR(hr))
		return false;

	// 3. Render
	if (!RenderCube())
		return false;

	// 4. End the scene and present it
	hr=m_d3dDevice->EndScene();
	if (!CheckHR(hr))
		return false;

	hr=m_d3dDevice->Present(NULL,NULL,NULL,NULL);
	if (!CheckHR(hr))
		return false;

	return true;
}

/*
To draw our geometry we need to call one of the D3D draw primitive functions
but first we need to specify our vertex buffer, index buffer and what FVF we are using
Since the world matrix is always set per object drawn we need to calculate and draw that as well
Here, in order to make the cube rotate differently, I maintain some static variables for rotation
*/
bool CGraphics::RenderCube()
{
	static bool rotateY=true;
	static int lastTime=timeGetTime();
	static float yRotation=0.0f;
	static float xRotation=0.0f;

	// Rotate the cube
	if (rotateY)
		yRotation+=0.05f;
	else
		xRotation+=0.05f;

	// After 4 seconds switch rotation axis
	if (timeGetTime()-lastTime>4000)
	{
		rotateY=!rotateY;
		lastTime=timeGetTime();
	}

	/* Create our world matrix by using the D3DX helper functions to define rotation 
	and translation matrices. The final world matrix is made by multiplying these together*/

	D3DXMATRIX matRotX,matRotY;
	D3DXMatrixRotationY( &matRotY, yRotation);
	D3DXMatrixRotationX( &matRotX, xRotation);

	D3DXMATRIX matTrans;
	D3DXMatrixTranslation(&matTrans,0,0,0);

	// Note: the order of multiplication here is very important
	D3DXMATRIX matWorld=(matRotY*matRotX)*matTrans;

	// Set all the values to render our cube
	m_d3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
	m_d3dDevice->SetMaterial( &m_material );
	m_d3dDevice->SetStreamSource( 0, m_vertexBuffer,0, sizeof(TCubeVertex) );
	m_d3dDevice->SetFVF( D3DFVF_CUBEVERTEX );
	m_d3dDevice->SetIndices( m_indexBuffer);

	// draw a triangle list using our 24 vertices and 12 triangles
	HRESULT hr=m_d3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST,0,0,24,0,12);  
	if (!CheckHR(hr))
		return false;

	return true;
}